package at.ac.uibk.gitsearch.security.jwt;

import at.ac.uibk.gitsearch.management.SecurityMetersService;
import io.jsonwebtoken.Claims;
import io.jsonwebtoken.ExpiredJwtException;
import io.jsonwebtoken.JwtBuilder;
import io.jsonwebtoken.JwtParser;
import io.jsonwebtoken.Jwts;
import io.jsonwebtoken.MalformedJwtException;
import io.jsonwebtoken.SignatureAlgorithm;
import io.jsonwebtoken.UnsupportedJwtException;
import io.jsonwebtoken.io.Decoders;
import io.jsonwebtoken.security.Keys;
import io.jsonwebtoken.security.SignatureException;
import java.nio.charset.StandardCharsets;
import java.security.Key;
import java.util.Arrays;
import java.util.Collection;
import java.util.Date;
import java.util.Map;
import java.util.Optional;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;
import org.apache.commons.lang3.ObjectUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.security.authentication.UsernamePasswordAuthenticationToken;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.GrantedAuthority;
import org.springframework.security.core.authority.SimpleGrantedAuthority;
import org.springframework.security.core.context.SecurityContext;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.security.core.userdetails.User;
import org.springframework.security.oauth2.core.oidc.user.DefaultOidcUser;
import org.springframework.stereotype.Component;
import tech.jhipster.config.JHipsterProperties;

@Component
public class TokenProvider {

    public static final String GITLAB_ACCESS_TOKEN = "gitLabAccessToken";
    public static final String GITLAB_ACCESS_ISSUER = "gitLabAccessIssuer";

    @Value("${application.gitlab.url}")
    private String gitLabUrl;

    public static final String PRE_TOKEN_CLAIM = "preToken";

    private final Logger log = LoggerFactory.getLogger(TokenProvider.class);

    private static final String AUTHORITIES_KEY = "auth";

    private static final String INVALID_JWT_TOKEN = "Invalid JWT token.";

    private final Key key;

    private final JwtParser jwtParser;

    private final long tokenValidityInMilliseconds;

    private final long tokenValidityInMillisecondsForRememberMe;

    private final SecurityMetersService securityMetersService;

    public TokenProvider(JHipsterProperties jHipsterProperties, SecurityMetersService securityMetersService) {
        byte[] keyBytes;
        String secret = jHipsterProperties.getSecurity().getAuthentication().getJwt().getBase64Secret();
        if (ObjectUtils.isEmpty(secret)) {
            log.warn(
                "Warning: the JWT key used is not Base64-encoded. " +
                "We recommend using the `jhipster.security.authentication.jwt.base64-secret` key for optimum security."
            );
            secret = jHipsterProperties.getSecurity().getAuthentication().getJwt().getSecret();
            keyBytes = secret.getBytes(StandardCharsets.UTF_8);
        } else {
            log.debug("Using a Base64-encoded JWT secret key");
            keyBytes = Decoders.BASE64.decode(secret);
        }
        key = Keys.hmacShaKeyFor(keyBytes);
        jwtParser = Jwts.parserBuilder().setSigningKey(key).build();
        this.tokenValidityInMilliseconds = 1000 * jHipsterProperties.getSecurity().getAuthentication().getJwt().getTokenValidityInSeconds();
        this.tokenValidityInMillisecondsForRememberMe =
            1000 * jHipsterProperties.getSecurity().getAuthentication().getJwt().getTokenValidityInSecondsForRememberMe();

        this.securityMetersService = securityMetersService;
    }

    /**
     * creates a token for authentication with validity determined by rememberMe.
     * @param authentication
     * @param rememberMe
     * @return
     */
    public String createToken(Authentication authentication, boolean rememberMe) {
        long validity = rememberMe ? tokenValidityInMillisecondsForRememberMe : tokenValidityInMilliseconds;

        return createToken(authentication, validity, false);
    }

    /**
     * creates a token from authentication given by validity (im msec)
     * @param authentication the authentication
     * @param validity validity in msec
     * @param isPreToken include hint that this token entitles for a long term token
     * @return
     */
    public String createToken(Authentication authentication, long validity, boolean isPreToken) {
        Date endTime = new Date(System.currentTimeMillis() + validity);

        String authorities = authentication.getAuthorities().stream().map(GrantedAuthority::getAuthority).collect(Collectors.joining(","));
        JwtBuilder jwtBuilder = Jwts.builder().setSubject(authentication.getName()).claim(AUTHORITIES_KEY, authorities);
        // copy from preToken
        String authenticationToken = null;
        String authenticationIssuer = null;
        if (authentication.getDetails() instanceof Map) {
            @SuppressWarnings("unchecked")
            Map<String, String> details = (Map<String, String>) authentication.getDetails();
            authenticationToken = details.get(GITLAB_ACCESS_TOKEN);
            authenticationIssuer = details.get(GITLAB_ACCESS_ISSUER);
        }
        if (authentication.getPrincipal() instanceof DefaultOidcUser) {
            DefaultOidcUser oidcInfo = (DefaultOidcUser) authentication.getPrincipal();
            authenticationToken = oidcInfo.getClaimAsString(GITLAB_ACCESS_TOKEN);
            authenticationIssuer = oidcInfo.getClaimAsString(GITLAB_ACCESS_ISSUER);
        }

        if (authenticationToken != null) {
            jwtBuilder.claim(GITLAB_ACCESS_TOKEN, authenticationToken);
            jwtBuilder.claim(GITLAB_ACCESS_ISSUER, authenticationIssuer);
        }

        if (isPreToken) {
            jwtBuilder = jwtBuilder.claim(PRE_TOKEN_CLAIM, PRE_TOKEN_CLAIM);
        }
        return jwtBuilder.signWith(key, SignatureAlgorithm.HS512).setExpiration(endTime).compact();
    }

    public Authentication getAuthentication(String token) {
        Claims claims = Jwts.parserBuilder().setSigningKey(key).build().parseClaimsJws(token).getBody();

        Collection<? extends GrantedAuthority> authorities = Arrays
            .stream(claims.get(AUTHORITIES_KEY).toString().split(","))
            .map(SimpleGrantedAuthority::new)
            .collect(Collectors.toList());
        String preTokenFlag = (String) claims.get(PRE_TOKEN_CLAIM);

        User principal = new User(claims.getSubject(), "", authorities);

        final UsernamePasswordAuthenticationToken authentication = new UsernamePasswordAuthenticationToken(principal, token, authorities);
        Map<String, String> detailsMap = new ConcurrentHashMap<>();
        if (preTokenFlag != null) {
            detailsMap.put(TokenProvider.PRE_TOKEN_CLAIM, preTokenFlag);
        }
        String authenticationToken = (String) claims.get(GITLAB_ACCESS_TOKEN);
        String authenticationIssuer = (String) claims.get(GITLAB_ACCESS_ISSUER);
        if (authenticationToken != null) {
            detailsMap.put(GITLAB_ACCESS_TOKEN, authenticationToken);
            detailsMap.put(GITLAB_ACCESS_ISSUER, authenticationIssuer);
        }
        authentication.setDetails(detailsMap);
        return authentication;
    }

    public boolean validateToken(String authToken) {
        try {
            jwtParser.parseClaimsJws(authToken);

            return true;
        } catch (ExpiredJwtException e) {
            this.securityMetersService.trackTokenExpired();

            log.trace(INVALID_JWT_TOKEN, e);
        } catch (UnsupportedJwtException e) {
            this.securityMetersService.trackTokenUnsupported();

            log.trace(INVALID_JWT_TOKEN, e);
        } catch (MalformedJwtException e) {
            this.securityMetersService.trackTokenMalformed();

            log.trace(INVALID_JWT_TOKEN, e);
        } catch (SignatureException e) {
            this.securityMetersService.trackTokenInvalidSignature();

            log.trace(INVALID_JWT_TOKEN, e);
        } catch (IllegalArgumentException e) { // TODO: should we let it bubble (no catch), to avoid defensive programming and follow the fail-fast principle?
            log.error("Token validation error {}", e.getMessage());
        }

        return false;
    }

    public Optional<User> getCurrentPrincipal() {
        SecurityContext securityContext = SecurityContextHolder.getContext();
        Authentication a = securityContext.getAuthentication();

        final Object principal = a.getPrincipal();

        if (principal instanceof User) {
            return Optional.ofNullable((User) principal);
        }
        return Optional.empty();
    }

    public Optional<String> getGitLabAccessToken() {
        SecurityContext securityContext = SecurityContextHolder.getContext();
        Authentication authentication = securityContext.getAuthentication();
        if (authentication == null) {
            return Optional.empty();
        }
        final Object userDetails = authentication.getDetails();
        if (userDetails instanceof Map<?, ?>) {
            @SuppressWarnings("unchecked")
            Map<String, String> userDetailsMap = (Map<String, String>) userDetails;
            return Optional.ofNullable(userDetailsMap.get(GITLAB_ACCESS_TOKEN));
        }
        return Optional.empty();
    }

    /**
     * returns the issuer URL for Gitlab API.
     * @return
     */
    public Optional<String> getGitLabAccessIssuer() {
        SecurityContext securityContext = SecurityContextHolder.getContext();
        Authentication authentication = securityContext.getAuthentication();
        final Object userDetails = authentication.getDetails();
        if (userDetails instanceof Map<?, ?>) {
            @SuppressWarnings("unchecked")
            Map<String, String> userDetailsMap = (Map<String, String>) userDetails;
            return Optional.ofNullable(userDetailsMap.get(GITLAB_ACCESS_ISSUER));
        }
        return Optional.empty();
    }

    public Optional<GitLabAccessInfo> getGitLabAccessInfo() {
        final Optional<String> accessToken = getGitLabAccessToken();
        if (accessToken.isPresent()) {
            final Optional<String> accessIssuer = getGitLabAccessIssuer();
            if (accessIssuer.isPresent()) {
                return Optional.of(new GitLabAccessInfo(accessToken.get(), gitLabUrl, getCurrentPrincipal()));
            }
        }
        return Optional.empty();
    }

    /**
     * encapsulate gitlab access token and access issuer
     * @author Michael Breu
     *
     */
    public static class GitLabAccessInfo {

        String accessToken;
        String gitLabAccessIssuer;
        Optional<User> user;

        public GitLabAccessInfo(String accessToken, String gitLabAccessIssuer, Optional<User> user) {
            super();
            this.accessToken = accessToken;
            this.gitLabAccessIssuer = gitLabAccessIssuer;
            this.user = user;
        }

        /**
         * @return the accessToken
         */
        public String getAccessToken() {
            return accessToken;
        }

        /**
         * @return the gitLabAccessIssuer
         */
        public String getGitLabAccessIssuer() {
            return gitLabAccessIssuer;
        }

        /**
         * @return the user
         */
        public Optional<User> getUser() {
            return user;
        }
    }
}
