package at.ac.uibk.gitsearch.service;

import at.ac.uibk.gitsearch.repository.gitlab.GitLabRepository;
import at.ac.uibk.gitsearch.repository.search.MetaDataRepository;
import at.ac.uibk.gitsearch.security.AuthoritiesConstants.AuthoritiesConstantEnum;
import at.ac.uibk.gitsearch.security.jwt.TokenProvider;
import at.ac.uibk.gitsearch.security.jwt.TokenProvider.GitLabAccessInfo;
import at.ac.uibk.gitsearch.service.dto.AutoCompleteEntry;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.nio.file.attribute.FileTime;
import java.text.ParseException;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipOutputStream;
import org.apache.commons.lang3.StringUtils;
import org.codeability.sharing.plugins.api.search.SearchInputDTO;
import org.codeability.sharing.plugins.api.search.SearchInputMetadataDTO;
import org.codeability.sharing.plugins.api.search.SearchResultDTO;
import org.codeability.sharing.plugins.api.search.SearchResultsDTO;
import org.codeability.sharing.plugins.api.search.util.ExerciseId;
import org.gitlab4j.api.GitLabApi;
import org.gitlab4j.api.GitLabApiException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.security.core.userdetails.User;
import org.springframework.stereotype.Service;
import org.springframework.web.servlet.support.ServletUriComponentsBuilder;

/**
 * Service for exercise/course search results
 * <p>
 */
@Service
@SuppressWarnings({ "PMD.TooManyMethods", "PMD.CognitiveComplexity" })
public class SearchService {

    private final Logger log = LoggerFactory.getLogger(SearchService.class);

    protected final MetaDataRepository metaDataRepository;
    protected final TokenProvider tokenProvider;
    protected final GitLabRepository gitLabRepository;
    protected final GitlabService gitlabService;
    protected final PluginManagementService pluginManagementService;
    protected final ShoppingBasketService shoppingBasketService;

    public SearchService(
        MetaDataRepository metaDataRepository,
        TokenProvider tokenProvider,
        GitLabRepository gitLabRepository,
        GitlabService gitlabService,
        PluginManagementService pluginManagementService,
        ShoppingBasketService shoppingBasketService
    ) {
        this.metaDataRepository = metaDataRepository;
        this.tokenProvider = tokenProvider;
        this.gitLabRepository = gitLabRepository;
        this.gitlabService = gitlabService;
        this.pluginManagementService = pluginManagementService;
        this.shoppingBasketService = shoppingBasketService;
    }

    /**
     * returns all keyword autocompletes for keyWord
     *
     * @param keyWordPrefix
     * @return
     * @throws IOException
     */
    public List<AutoCompleteEntry> getKeywordsAutoComplete(String keyWordPrefix, int max) throws IOException {
        return metaDataRepository.getKeywordsAutoComplete(keyWordPrefix, max);
    }

    /**
     * returns all keyword autocompletes for keyWord
     *
     * @param formatPrefix
     * @return
     * @throws IOException
     */
    public List<AutoCompleteEntry> getFormatsAutoComplete(String formatPrefix, int max) throws IOException {
        return metaDataRepository.getFormatsAutoComplete(formatPrefix, max);
    }

    /**
     * returns all creator autocompletes
     *
     * @param creatorPrefix
     * @return
     * @throws IOException
     */
    public List<AutoCompleteEntry> getCreatorAutoComplete(String creatorPrefix, int max) throws IOException {
        return metaDataRepository.getCreatorAutoComplete(creatorPrefix, max);
    }

    /**
     * returns all contributor autocompletes
     *
     * @param contributorPrefix
     * @return
     * @throws IOException
     */
    public List<AutoCompleteEntry> getContributorAutoComplete(String contributorPrefix, int max) throws IOException {
        return metaDataRepository.getContributorAutoComplete(contributorPrefix, max);
    }

    /**
     * returns all contributor and author autocompletes
     *
     * @param contributorPrefix
     * @return
     * @throws IOException
     */
    public List<AutoCompleteEntry> getContributorCreatorAutoComplete(String contributorPrefix, int max) throws IOException {
        return metaDataRepository.getContributorCreatorAutoComplete(contributorPrefix, max);
    }

    /**
     * returns all programmingLanguage autocompletes for keyWord
     *
     * @param programmingLanguagePrefix
     * @return
     * @throws IOException
     */
    public List<AutoCompleteEntry> getProgrammingLanguageAutoComplete(String programmingLanguagePrefix, int max) throws IOException {
        return metaDataRepository.getProgrammingLanguageAutoComplete(programmingLanguagePrefix, max);
    }

    /**
     * returns the result page
     *
     * @param searchInput the query definition
     */
    public SearchResultsDTO searchResultPage(SearchInputDTO searchInput) throws IOException {
        final Optional<User> principal = tokenProvider.getCurrentPrincipal();
        log.debug("Searchrequest for {} ", searchInput);

        final SearchResultsDTO pageDetails = metaDataRepository.pageDetails(searchInput, principal);

        pageDetails
            .getSearchResult()
            .stream()
            .forEach(hit ->
                pluginManagementService
                    .getRegisteredPluginConfigs()
                    .stream()
                    .forEach(config ->
                        config
                            .getActions()
                            .values()
                            .stream()
                            .filter(action -> action.isApplicable(hit))
                            .forEach(action -> hit.getSupportedActions().add(action.getPluginActionInfo()))
                    )
            );
        pageDetails.getSearchResult().stream().forEach(this::fixImageURL);
        pageDetails.getSearchResult().forEach(this::filterArtemisActionForUnauthorized);

        return pageDetails;
    }

    public SearchResultsDTO searchResultForAuthorEmail(String searchInput, int length) throws IOException {
        log.debug("Searchrequest for {} ", searchInput);

        final SearchResultsDTO pageDetails = metaDataRepository.searchByEmail(searchInput, length, tokenProvider.getCurrentPrincipal());

        pageDetails
            .getSearchResult()
            .stream()
            .forEach(hit ->
                pluginManagementService
                    .getRegisteredPluginConfigs()
                    .stream()
                    .forEach(config ->
                        config
                            .getActions()
                            .values()
                            .stream()
                            .filter(action -> action.isApplicable(hit))
                            .forEach(action -> hit.getSupportedActions().add(action.getPluginActionInfo()))
                    )
            );
        pageDetails.getSearchResult().stream().forEach(this::fixImageURL);

        return pageDetails;
    }

    /**
     * returns the result page
     *
     * @param page        the index of the first record to be returned
     * @param length      the number of records on page
     */
    public SearchResultsDTO watchListResultPage(Stream<String> exerciseIds, int page, int length) throws IOException {
        final SearchResultsDTO pageDetails = metaDataRepository.getExercisesById(
            exerciseIds,
            tokenProvider.getCurrentPrincipal(),
            page,
            length
        );

        pageDetails
            .getSearchResult()
            .stream()
            .forEach(hit ->
                pluginManagementService
                    .getRegisteredPluginConfigs()
                    .stream()
                    .forEach(config ->
                        config
                            .getActions()
                            .values()
                            .stream()
                            .filter(action -> action.isApplicable(hit))
                            .forEach(action -> hit.getSupportedActions().add(action.getPluginActionInfo()))
                    )
            );
        pageDetails.getSearchResult().stream().forEach(this::fixImageURL);

        return pageDetails;
        // return readTestResults(searchInput.getFulltextQuery(), first, length);
    }

    private void fixImageURL(SearchResultDTO metaData) {
        String image = metaData.getMetadata().getImage();

        try {
            if (image != null) {
                URI url = new URI(image);
                if (url.isAbsolute()) {
                    return;
                }
                String httpUrlToRepo = metaData.getProject().getUrl();
                String commitId = metaData.getFile().getCommit_id();
                String baseRepoURL = httpUrlToRepo + "/-/raw/" + commitId + "/";
                final URI resolvedImageUrl = new URI(baseRepoURL).resolve(url);
                metaData.getMetadata().setImage(resolvedImageUrl.toASCIIString());
                return;
            }

            final String baseUrl = ServletUriComponentsBuilder.fromCurrentContextPath().build().toUriString();
            final boolean isJava =
                metaData.getMetadata().getProgrammingLanguage() != null &&
                Arrays.stream(metaData.getMetadata().getProgrammingLanguage()).anyMatch(s -> s.startsWith("JAVA")) &&
                metaData.hashCode() % 2 == 0; // nur um ein wenig Abwechslung zu bekommen :-)
            if (isJava) {
                metaData.getMetadata().setImage(baseUrl + "/content/img/java.png");
                return;
            }
            // depends on code commented out above
            //			if (isLatexFormat) {
            //				metaData.getMetadata().setImage(baseUrl + "/content/img/latex.png");
            //				return;
            //			}
            final boolean isPython =
                metaData.getMetadata().getProgrammingLanguage() != null &&
                Arrays.stream(metaData.getMetadata().getProgrammingLanguage()).anyMatch(s -> s.startsWith("python"));
            if (isPython) {
                metaData.getMetadata().setImage(baseUrl + "/content/img/python.png");
                return;
            }
            metaData.getMetadata().setImage(baseUrl + "/content/img/gitLab.png");
            return;
        } catch (URISyntaxException e) {
            log.warn("Cannot parse image url {} for {}", metaData.getMetadata().getImage(), metaData.getProject().getUrl());
        }
    }

    public File exportExercise(String exerciseId) throws IOException, ParseException {
        return exportExercise(ExerciseId.fromString(exerciseId), ExtractionDepth.JUST_PROJECT);
    }

    public enum ExtractionDepth {
        JUST_PROJECT,
        WITH_CHILDREN,
        WITH_DESCENDANTS,
    }

    @SuppressWarnings({ "PMD.CloseResource", "PMD.AvoidDuplicateLiterals" })
    public File exportExercise(ExerciseId exerciseId, ExtractionDepth recursive) throws IOException {
        ZipInputStream inputFile = getZippedExercise(exerciseId, tokenProvider.getCurrentPrincipal(), recursive, new HashSet<>());
        if (inputFile == null) {
            return null;
        }
        File file = File.createTempFile("exercise" + exerciseId.getProjectId(), ".zip");

        return copyInputStreamToFile(inputFile, file);
    }

    @SuppressWarnings({ "PMD.CloseResource", "PMD.AvoidDuplicateLiterals" })
    private ZipInputStream getZippedExercise(
        ExerciseId exerciseId,
        Optional<User> user,
        ExtractionDepth recursive,
        Set<ExerciseId> alreadyTreatedExerciseIds
    ) throws IOException {
        try {
            final SearchResultDTO exercise = metaDataRepository.getExerciseById(exerciseId.toString(), user);
            final GitLabApi gitLabApi = gitLabRepository.getGitLabApi(tokenProvider.getGitLabAccessInfo());
            // first we try with user credentials!
            InputStream repositoryArchive;
            String[] filter;
            try {
                repositoryArchive = gitLabApi.getRepositoryApi().getRepositoryArchive(exercise.getProject().getProject_id(), "HEAD", "zip");
                // if the previous gitlab requests succeeds, the user has full access to the repository.
                // no filtering is required
                filter = new String[] {};
            } catch (GitLabApiException e) {
                // if the previous gitlab throws an exception, we expect, that this is because of permission problems.
                // we check for public visibility, and construct the filter from the except array.
                log.warn("no access?", e.getMessage());
                if (exercise.getMetadata().getPublicVisibility() == null) {
                    throw e; // throw it further
                } else {
                    repositoryArchive =
                        gitLabRepository
                            .getAdminGitLabApi()
                            .getRepositoryApi()
                            .getRepositoryArchive(exercise.getProject().getProject_id(), "HEAD", "zip");
                    filter = exercise.getMetadata().getPublicVisibility().getExcept();
                }
            }
            InputStream project = shoppingBasketService.rePackageGitLabProjectZip(
                new ZipInputStream(repositoryArchive),
                "from project " + exerciseId.getProjectId(),
                filter,
                exerciseId.getPath()
            );
            if (recursive == ExtractionDepth.JUST_PROJECT) {
                return new ZipInputStream(project);
            } else {
                alreadyTreatedExerciseIds.add(exerciseId); // Side effects are always dangerous :-(
                return extendZipByChildren(exerciseId, new ZipInputStream(project), user, recursive, alreadyTreatedExerciseIds);
            }
        } catch (GitLabApiException exception) {
            log.error(exception.getMessage());
            return null;
        }
    }

    /**
     *
     * @param exerciseId the exercise id
     * @param existingZip a zip input stream that should be amended by children, may be null
     * @param user the user to check for permissions
     * @param recursive the recursion type
     * @param alreadyTreatedExerciseIds list of already treated Exercise Ids (in order to avoid endless recursions)
     * @return an zipped output stream containing also children
     * @throws IOException
     */
    @SuppressWarnings("PMD.CloseResource")
    private ZipInputStream extendZipByChildren(
        ExerciseId exerciseId,
        ZipInputStream existingZip,
        Optional<User> user,
        ExtractionDepth recursive,
        Set<ExerciseId> alreadyTreatedExerciseIds
    ) throws IOException {
        if (recursive == ExtractionDepth.JUST_PROJECT) {
            return existingZip;
        }

        SearchResultDTO exerciseMD = metaDataRepository.getExerciseById(exerciseId, user);
        ExtractionDepth childRecursion = (recursive == ExtractionDepth.WITH_CHILDREN)
            ? ExtractionDepth.JUST_PROJECT
            : ExtractionDepth.WITH_DESCENDANTS;
        if (exerciseMD != null && exerciseMD.getFile().getChildren() != null) {
            File file = File.createTempFile("exercise" + exerciseId.getProjectId() + "recursive", "zip");
            ZipOutputStream zos = new ZipOutputStream(new FileOutputStream(file));
            if (existingZip != null) {
                copyZip(existingZip, zos, null);
            }

            for (String child : exerciseMD.getFile().getChildren()) {
                try {
                    ExerciseId childExerciseId = ExerciseId.fromString(child);
                    if (childExerciseId.isDescendantOf(exerciseId)) {
                        ZipInputStream zis = extendZipByChildren(childExerciseId, null, user, childRecursion, alreadyTreatedExerciseIds);
                        copyZip(zis, zos, child);
                    } else {
                        ZipInputStream zis = getZippedExercise(childExerciseId, user, childRecursion, alreadyTreatedExerciseIds);
                        copyZip(zis, zos, child);
                    }
                } catch (IOException | ParseException e) {
                    log.error("Cannot extract children {}", child);
                }
            }
            zos.close();
            return new ZipInputStream(new FileInputStream(file));
        }
        // nothing copied
        return existingZip;
    }

    /**
     * copies a zipped input stream into the existing zipped output stream.
     * @param zis
     * @param zos
     * @param parent
     * @throws IOException
     */
    private void copyZip(ZipInputStream zis, ZipOutputStream zos, String parent) throws IOException {
        for (ZipEntry ze = zis.getNextEntry(); ze != null; ze = zis.getNextEntry()) {
            String newName = StringUtils.isEmpty(parent) ? ze.getName() : parent + "/" + ze.getName();
            ZipEntry newZE = new ZipEntry(newName);
            newZE.setComment(ze.getComment());
            FileTime creationTime = ze.getCreationTime();
            if (creationTime != null) {
                newZE.setCreationTime(creationTime);
            }

            zos.putNextEntry(newZE);
            byte[] buffer = new byte[1024];
            for (int read = zis.read(buffer); read != -1; read = zis.read(buffer)) {
                zos.write(buffer, 0, read);
            }
            zos.closeEntry();
        }

        zis.close();
    }

    private File copyInputStreamToFile(ZipInputStream inputStream, File file) throws IOException {
        // append = false
        try (ZipOutputStream outputStream = new ZipOutputStream(new FileOutputStream(file, false))) {
            copyZip(inputStream, outputStream, null);
            return file;
        }
    }

    public Boolean hasAccessToExerciseId(String exerciseId) {
        if (tokenProvider.getCurrentPrincipal().isEmpty()) {
            log.warn("Cannot find a principal for for exercise {}", exerciseId);
            return false;
        }
        if (tokenProvider.getCurrentPrincipal().get().getAuthorities().contains(AuthoritiesConstantEnum.ADMIN.getGrantedAuthority())) {
            log.info("Admin can access everything : {}", exerciseId);
            return true; // ADMIN is always allowed
        }
        SearchResultDTO result = metaDataRepository.getExerciseById(exerciseId, tokenProvider.getCurrentPrincipal());
        if (result == null) {
            log.warn("Cannot find exercise for : {}", exerciseId);
            return false;
        }
        return true;
    }

    public Optional<SearchResultDTO> findExerciseById(ExerciseId exerciseId) {
        try {
            SearchResultDTO result = metaDataRepository.getExerciseById(exerciseId, tokenProvider.getCurrentPrincipal());
            return Optional.ofNullable(result);
        } catch (javax.ws.rs.NotFoundException e) {
            log.error("exercise with id {} not found?", exerciseId, e);
            return Optional.empty();
        }
    }

    /**
     * Method used to filter out the Artemis export action for
     * non-authorized users having only partial access to the data.
     *
     * See Metadata entry 'publicVisibility.except'
     * @param result SearchResultDTO to filter the actions from
     * @return the updated SearchResultDTO with filtered actions
     */
    @SuppressWarnings({ "PMD.CloseResource", "PMD.AvoidCatchingGenericException", "PMD.AvoidCatchingNPE" })
    private SearchResultDTO filterArtemisActionForUnauthorized(SearchResultDTO result) {
        if (result.getMetadata().getPublicVisibility() == null || result.getMetadata().getPublicVisibility().getExcept() == null) {
            // results queried without 'publicVisibility' are already authorized
            return result;
        }
        try {
            final Optional<GitLabAccessInfo> gitLabAccessInfo = tokenProvider.getGitLabAccessInfo();
            GitLabApi gitLabApi = gitLabRepository.getGitLabApi(gitLabAccessInfo);
            gitLabApi.getRepositoryApi().getRepositoryArchive(result.getProject().getProject_id(), "HEAD", "zip");
        } catch (GitLabApiException | NullPointerException e) { // TODO NullPointerException thrown if gitLabAccessInfo.isEmpty(): make it more elegant
            // We expect permissions missing
            result.setSupportedActions(
                result
                    .getSupportedActions()
                    .stream()
                    .filter(action -> !"Artemis Sharing Connector".equals(action.getPlugin()))
                    .collect(Collectors.toList())
            );
        }
        return result;
    }

    /**
     * Method to search for all public exercises.
     * Searches all public exercises with an empty search query and without a specified user.
     *
     * @return the search result page DTO
     * @throws IOException
     */
    public SearchResultsDTO searchAllPublicResults() throws IOException {
        log.debug("Search request to find all public exercises");

        final SearchInputMetadataDTO emptySearchMetadata = new SearchInputMetadataDTO(null, null, null, null, null, null);
        SearchInputDTO emptySearchQuery = new SearchInputDTO(null, emptySearchMetadata, null, null, null, 0);

        return metaDataRepository.pageDetails(emptySearchQuery, Optional.empty());
    }
}
